{$G+,N+}                                                { 286+, 287+ }
{$R-,S-}
Unit MapIO;

Interface

Uses Grafix;

Type
  TVertex = Record
    X, Y : LongInt
  End;

  TLineDef = Record
    Typ,
    Vertex1, Vertex2,
    Owner, Texture, PLineDef, PSector : Integer;
    Len : Double;
  End;

  TSector = Record
    SLineDef, ELineDef,
    FHeight,  CHeight,
    FTexture, CTexture : Integer;
  End;

  TTexture = Record
    Size : Word;
    P : Pointer;
  End;

Const
  Max_Vertices = 65528 div SizeOf(TVertex);
  Max_LineDefs = 65528 div SizeOf(TLineDef);
  Max_Sectors  = 65528 div SizeOf(TSector);
  Max_Textures = 65528 div SizeOf(TTexture);

Type
  TVertices = Array [0..Max_Vertices-1] of TVertex;
  TLineDefs = Array [0..Max_LineDefs-1] of TLineDef;
  TSectors  = Array [0..Max_Sectors-1]  of TSector;
  TTextures = Array [0..Max_Textures-1] of TTexture;

Var
  Vertices_Cnt,
  LineDefs_Cnt,
  Sectors_Cnt,
  Textures_Cnt : Integer;
  Palette : TPal;                               { Color palette }

  PVertices : ^TVertices;
  PLineDefs : ^TLineDefs;
  PSectors  : ^TSectors;
  PTextures: ^TTextures;

  Sector : Integer;
  Xp, Yp, Zp : Double;
  Angle,                                { Angle for looking around }
  YAng : Integer;                       { Angle for looking up/down }
{ Temporary player coords & angles }
  _Sector : Integer;
  _Xp, _Yp, _Zp : Double;
  _Angle,                                { Angle for looking around }
  _YAng : Integer;                       { Angle for looking up/down }


Function Load_Map (Name : String) : Boolean;
Procedure UnLoad_Map;

Implementation

Procedure Del_Heading_Spaces (Var S : String);
Begin
  If S = '' then Exit;
  While S[1] = ' ' do Delete(S,1,1)
End;

Function Scan_Str (S, Param : String; Cnt : Word;
                   Var Values : Array of LongInt) : Boolean;
Var
  Index : Integer;
  Result : Boolean;
  I : Word;
  SS : String;
  Err : Integer;
Label
  L, LEnd;
Begin
  Result := False;
  If (Cnt = 0) or (S = '') or (Param = '') then Goto LEnd;
  Index := -1;
  I := 1; While (S[I] = ' ') and (I <= Length(S)) do Inc (I);
  If I > Length(S) then Goto LEnd;
  SS := '';
  While I <= Length(S) do Begin
    If S[I] <> ' ' then Begin
      SS := SS+S[I];
      Inc (I);
      If I > Length(S) then Goto L
    End
    Else Begin
L:
      If Index < 0 then Begin
        If SS <> Param then Goto LEnd;
        Inc (Index)
      End
      Else Begin
        Val (SS, Values[Index], Err);
        If Err <> 0 then Goto LEnd;
        Inc (Index);
        If Index = Cnt then Break
      End;
      SS := '';
      While (S[I] = ' ') and (I <= Length(S)) do Inc (I)
    End
  End;
  If Index = Cnt then Result := True;
LEnd:
  Scan_Str := Result
End;

Function Load_Map (Name : String) : Boolean;
Var
  T : Text;
  S : String;
  Result : Boolean;
  I, J, Phase : Integer;
  Values : Array [0..15] of LongInt;
  Scale_Factor : LongInt;
  DX, DY : Double;
  TName : String;
Label
  LEnd, LClose;
Begin
  PVertices := Nil;
  PLineDefs := Nil;
  PSectors := Nil;
  PTextures := Nil;
  Result := False;
  Assign (T, Name);
  {$I-} Reset (T); {$I+}
  If IOResult <> 0 then Goto LEnd;

  Phase := 1;
  While not EOF(T) do Begin
    ReadLn (T, S);
    Del_Heading_Spaces (S);
    If (S = '') or (Pos('//',S)=1) then Continue; { Skip empty ones & comments }

    Case Phase of
      1:
          Begin
            If not Scan_Str (S, 'max_vertices', 1, Values) then Goto LClose;
            Vertices_Cnt := Values[0];
            Inc (Phase)
          End;
      2:
          Begin
            If not Scan_Str (S, 'max_linedefs', 1, Values) then Goto LClose;
            LineDefs_Cnt := Values[0];
            Inc (Phase)
          End;
      3:
          Begin
            If not Scan_Str (S, 'max_sectors', 1, Values) then Goto LClose;
            Sectors_Cnt := Values[0];
            Inc (Phase)
          End;
      4:
          Begin
            If not Scan_Str (S, 'max_textures', 1, Values) then Goto LClose;
            Textures_Cnt := Values[0];
            Inc (Phase)
          End;
      5:
          Begin
            If not Scan_Str (S, 'scale_factor', 1, Values) then Goto LClose;
            Scale_Factor := Values[0];
            Inc (Phase)
          End;
      6:
          Begin
            If not Scan_Str (S, 'player_pos', 3, Values) then Goto LClose;
            Xp := Values[0]*Scale_Factor; Yp := Values[1]*Scale_Factor;
            Angle := Values[2];
            Inc (Phase);
            If MaxAvail < Vertices_Cnt*SizeOf(TVertex) then Goto LClose;
            GetMem (PVertices, Vertices_Cnt*SizeOf(TVertex));
            I := 0;
          End;
      7:
          Begin
            If not Scan_Str (S, 'vertex', 3, Values) then Goto LClose;
            with PVertices^[I] do Begin
              X := Values[1]*Scale_Factor;
              Y := Values[2]*Scale_Factor
            End;
            Inc (I);
            If I = Vertices_Cnt then Begin
              Inc (Phase);
              If MaxAvail < LineDefs_Cnt*SizeOf(TLineDef) then Goto LClose;
              GetMem (PLineDefs, LineDefs_Cnt*SizeOf(TLineDef));
              I := 0
            End
          End;
      8:
          Begin
            If not Scan_Str (S, 'linedef', 5, Values) then Goto LClose;
            with PLineDefs^[I] do Begin
              If Values[3] <> -1 then Typ := 1 Else Typ := 0;
              Vertex1 := Values[1];
              Vertex2 := Values[2];
              PLineDef := Values[3];
              Texture := Values[4];
              DX := PVertices^[Vertex2].X - PVertices^[Vertex1].X;
              DY := PVertices^[Vertex2].Y - PVertices^[Vertex1].Y;
              Len := Sqrt (Sqr(DX) + Sqr(DY))
            End;
            Inc (I);
            If I = LineDefs_Cnt then Begin
              Inc (Phase);
              If MaxAvail < Sectors_Cnt*SizeOf(TSector) then Goto LClose;
              GetMem (PSectors, Sectors_Cnt*SizeOf(TSector));
              I := 0
            End
          End;
      9:
          Begin
            If not Scan_Str (S, 'sector', 7, Values) then Goto LClose;
            with PSectors^[I] do Begin
              SLineDef := Values[1];
              ELineDef := Values[2];
              FHeight := Values[3];
              CHeight := Values[4];
              FTexture := Values[5];
              CTexture := Values[6]
            End;
            Inc (I);
            If I = Sectors_Cnt then Begin
              Inc (Phase);
              If Textures_Cnt <> 0 then Begin
                If MaxAvail < Textures_Cnt*SizeOf(TTexture) then Goto LClose;
                GetMem (PTextures, Textures_Cnt*SizeOf(TTexture));
                FillChar (PTextures^, Textures_Cnt*SizeOf(TTexture), 0)
              End
              Else
                Inc (Phase);
              I := 0
            End
          End;
      10:
          Begin
            If not Scan_Str (S, 'texture', 2, Values) then Goto LClose;
            Str (Values[1], TName);
            TName := TName + '.PCX';
            LoadPCX256 (TName, PTextures^[I].P, PTextures^[I].Size, Palette);
            If PTextures^[I].Size = 0 then Goto LClose;
            Inc (I);
            If I = Textures_Cnt then Begin
              Inc (Phase);
              { ... }
              I := 0
            End
          End
    End;

    If Phase = 11 then Begin
      Result := True;
      Break
    End
  End;

LClose:
  Close (T);

LEnd:
  If Result then Begin
    For I := 0 to LineDefs_Cnt-1 do
      For J := 0 to Sectors_Cnt-1 do
        If (PSectors^[J].SLineDef <= I) and (I <= PSectors^[J].ELineDef) then
          PLineDefs^[I].Owner := J;

    For I := 0 to LineDefs_Cnt-1 do
      If PLineDefs^[I].Typ and 1 = 1 then
        PLineDefs^[I].PSector := PLineDefs^[PLineDefs^[I].PLineDef].Owner
      Else
        PLineDefs^[I].PSector := -1
  End
  Else Begin
    If PVertices <> Nil then FreeMem (PVertices, Vertices_Cnt*SizeOf(TVertex));
    If PLineDefs <> Nil then FreeMem (PLineDefs, LineDefs_Cnt*SizeOf(TLineDef));
    If PSectors  <> Nil then FreeMem (PSectors,  Sectors_Cnt *SizeOf(TSector));
    If PTextures <> Nil then Begin
      For I := 0 to Textures_Cnt-1 do
        If PTextures^[I].Size <> 0 then
          FreeMem (PTextures^[I].P, PTextures^[I].Size);
      FreeMem (PTextures, Textures_Cnt*SizeOf(TTexture))
    End;
    PVertices := Nil;
    PLineDefs := Nil;
    PSectors := Nil;
    PTextures := Nil
  End;

  Load_Map := Result
End;

Procedure UnLoad_Map;
Var
  I : Integer;
Begin
  If PVertices <> Nil then
    FreeMem (PVertices, Vertices_Cnt*SizeOf(TVertex));
  If PLineDefs <> Nil then
    FreeMem (PLineDefs, LineDefs_Cnt*SizeOf(TLineDef));
  If PSectors <> Nil then
    FreeMem (PSectors,  Sectors_Cnt *SizeOf(TSector));
  If PTextures <> Nil then Begin
    For I := 0 to Textures_Cnt-1 do
      If PTextures^[I].Size <> 0 then
        FreeMem (PTextures^[I].P, PTextures^[I].Size);
    FreeMem (PTextures, Textures_Cnt*SizeOf(TTexture))
  End;
  PVertices := Nil;
  PLineDefs := Nil;
  PSectors := Nil;
  PTextures := Nil
End;

Begin
  PVertices := Nil;
  PLineDefs := Nil;
  PSectors := Nil;
  PTextures := Nil
End.
